@@ -352,15 +352,10 @@ expand_macros(const char *str, int except_dollar)
 /*
  * Process a non-command line
  */
-static char *
+static void
 process_line(char *s)
 {
-	char *r, *t;
-
-	// Skip leading blanks
-	while (isblank(*s))
-		s++;
-	r = s;
+	char *t;
 
 	// Strip comment
 #if ENABLE_FEATURE_MAKE_EXTENSIONS
@@ -395,8 +390,6 @@ process_line(char *s)
 		}
 	}
 	*t = '\0';
-
-	return r;
 }
 
 #if ENABLE_FEATURE_MAKE_EXTENSIONS
@@ -479,11 +472,8 @@ skip_line(const char *str1)
 	// Default is to return skip flag for current level
 	int ret = cstate[clevel] & SKIP_LINE;
 
-	if (*str1 == '\t')
-		return ret;
-
-	copy = xstrdup(str1);
-	q = process_line(copy);
+	q = copy = xstrdup(str1);
+	process_line(copy);
 	if ((token = gettok(&q)) != NULL) {
 		if (strcmp(token, "endif") == 0) {
 			if (gettok(&q) != NULL)
@@ -588,7 +578,7 @@ make_fgets(char *s, int size, FILE *fd)
  * Ignore comment lines.  Return NULL on EOF.
  */
 static char *
-readline(FILE *fd)
+readline(FILE *fd, int want_command)
 {
 	char *p, *str = NULL;
 	int pos = 0;
@@ -627,15 +617,21 @@ readline(FILE *fd)
 		}
 		dispno = lineno;
 
-		// Check for comment lines and lines that are conditionally skipped.
-		p = str;
-		while (isblank(*p))
-			p++;
+#if ENABLE_FEATURE_MAKE_EXTENSIONS
+		// Check for lines that are conditionally skipped.
+		if (posix || !skip_line(str))
+#endif
+		{
+			if (want_command && *str == '\t')
+				return str;
 
-		if (*p != '\n' && *str != '#'
-				IF_FEATURE_MAKE_EXTENSIONS(&& (posix || !skip_line(str)))
-		) {
-			return str;
+			// Check for comment lines
+			p = str;
+			while (isblank(*p))
+				p++;
+
+			if (*p != '\n' && *str != '#')
+				return str;
 		}
 
 		pos = 0;
@@ -903,6 +899,21 @@ wildcard(char *p, glob_t *gd)
 	}
 	return 1;
 }
+
+void
+pragmas_from_env(void)
+{
+	char *p, *q, *var;
+	const char *env = getenv("PDPMAKE_PRAGMAS");
+
+	if (env == NULL)
+		return;
+
+	q = var = xstrdup(env);
+	while ((p = gettok(&q)) != NULL)
+		set_pragma(p);
+	free(var);
+}
 #endif
 
 /*
@@ -935,11 +946,9 @@ input(FILE *fd, int ilevel)
 #endif
 
 	lineno = 0;
-	str1 = readline(fd);
+	str1 = readline(fd, FALSE);
 	while (str1) {
 		str2 = NULL;
-		if (*str1 == '\t')	// Command without target
-			error("command not allowed here");
 
 		// Newlines and comments are handled differently in command lines
 		// and other types of line.  Take a copy of the current line before
@@ -949,9 +958,15 @@ input(FILE *fd, int ilevel)
 		//   target: prereq; command
 		//
 		copy = xstrdup(str1);
-		str = process_line(str1);
+		process_line(str1);
+		str = str1;
 
 		// Check for an include line
+# if ENABLE_FEATURE_MAKE_EXTENSIONS
+		if (!posix)
+			while (isblank(*str))
+				++str;
+#endif
 #if ENABLE_FEATURE_MAKE_POSIX_2024
 		minus = !POSIX_2017 && *str == '-';
 #endif
@@ -963,11 +978,15 @@ input(FILE *fd, int ilevel)
 			if (ilevel > 16)
 				error("too many includes");
 
+#if ENABLE_FEATURE_MAKE_POSIX_2024
+			count = 0;
+#endif
 			q = expanded = expand_macros(p + 7, FALSE);
 			while ((p = gettok(&q)) != NULL) {
 				FILE *ifd;
 
 #if ENABLE_FEATURE_MAKE_POSIX_2024
+				++count;
 				if (!POSIX_2017) {
 					// Try to create include file or bring it up-to-date
 					opts |= OPT_include;
@@ -982,6 +1001,8 @@ input(FILE *fd, int ilevel)
 					makefile = p;
 					input(ifd, ilevel + 1);
 					fclose(ifd);
+					makefile = old_makefile;
+					lineno = old_lineno;
 				}
 #if ENABLE_FEATURE_MAKE_POSIX_2024
 				if (POSIX_2017)
@@ -995,7 +1016,7 @@ input(FILE *fd, int ilevel)
 				if (p == NULL || gettok(&q)) {
 					error("one include file per line");
 				}
-			} else if (makefile == old_makefile) {
+			} else if (count == 0) {
 				// In POSIX 2024 no include file is unspecified behaviour.
 # if ENABLE_FEATURE_MAKE_EXTENSIONS
 				if (posix)
@@ -1003,13 +1024,17 @@ input(FILE *fd, int ilevel)
 					error("no include file");
 			}
 #endif
-
-			makefile = old_makefile;
-			lineno = old_lineno;
 			goto end_loop;
 		}
 
 		// Check for a macro definition
+		str = str1;
+#if ENABLE_FEATURE_MAKE_EXTENSIONS || ENABLE_FEATURE_MAKE_POSIX_2024
+		// POSIX 2024 seems to allow a tab as the first character of
+		// a macro definition, though most implementations don't.
+		if (POSIX_2017 && *str == '\t')
+			error("command not allowed here");
+#endif
 		if (find_char(str, '=') != NULL) {
 			int level = (useenv || fd == NULL) ? 4 : 3;
 			// Use a copy of the line:  we might need the original
@@ -1132,6 +1157,8 @@ input(FILE *fd, int ilevel)
 
 		// If we get here it must be a target rule
  try_target:
+		if (*str == '\t')	// Command without target
+			error("command not allowed here");
 		p = expanded = expand_macros(str, FALSE);
 
 		// Look for colon separator
@@ -1227,7 +1254,7 @@ input(FILE *fd, int ilevel)
 
 		// Create list of commands
 		startno = dispno;
-		while ((str2 = readline(fd)) && *str2 == '\t') {
+		while ((str2 = readline(fd, TRUE)) && *str2 == '\t') {
 			cp = newcmd(process_command(str2), cp);
 			free(str2);
 		}
@@ -1286,7 +1313,7 @@ input(FILE *fd, int ilevel)
  end_loop:
 		free(str1);
 		dispno = lineno;
-		str1 = str2 ? str2 : readline(fd);
+		str1 = str2 ? str2 : readline(fd, FALSE);
 		free(copy);
 		free(expanded);
 #if ENABLE_FEATURE_MAKE_EXTENSIONS
